import argparse
import json
import locale
import re
import logging
from datetime import datetime, timedelta
import calendar

from pygments import highlight, lexers, formatters
from scalpl import Cut

from tww.localization import setlocale, resolve_locale
from tww.lib import ISO_FORMAT, time_to_emoji, time_ago, workday_diff, workhours_diff, td_remainders, td_totals, \
    td_iso8601, ISO_Z_FORMAT, td_human, td_machine, tzinfo_from_offset, td_human_months_weeks
from tww.lib import resolve_timezone, dateparser_parse_dt, get_utcnow, get_s_since_epoch, get_ms_since_epoch, \
    dt_tz_translation, get_local_now, query_to_format_result
from tww.common import logger

custom_locale = resolve_locale()
custom_tz = None

r_generic = re.compile('(.*)', flags=re.IGNORECASE)
r_time_in_epoch_s_now = re.compile('(?:time since epoch|seconds since epoch)', flags=re.IGNORECASE)
r_time_in_epoch_s2 = re.compile('(.*)?\s*(?:in|to)\s*(?:epoch|seconds since epoch|seconds)', flags=re.IGNORECASE)
r_time_in_epoch_s3 = re.compile('(?:seconds)?\s*since\s*(.*)', flags=re.IGNORECASE)
r_time_in_epoch_ms_now = re.compile('(?:milliseconds since epoch)', flags=re.IGNORECASE)
r_time_in_epoch_ms2 = re.compile('(.*)?\s*(?:in|to)\s*(?:ms|milliseconds|miliseconds)', flags=re.IGNORECASE)
r_time_in_epoch_ms3 = re.compile('(?:ms|milliseconds|miliseconds)?\s*since\s*(.*)', flags=re.IGNORECASE)
r_time_in = re.compile('(?:time)?\s*in\s*(.*)', flags=re.IGNORECASE)
r_time_since = re.compile('(?:time|year|month|week|day|hour|minute|second)?(?:s)?\s*(?:since|ago)\s*(.*)', flags=re.IGNORECASE)
r_time_until = re.compile('(?:time|year|month|week|day|hour|minute|second)?(?:s)?\s*(?:until|to)\s*(.*)', flags=re.IGNORECASE)
r_time_between = re.compile('(?:time|year|month|week|day|hour|minute|second)?(?:s)?\s*between\s*(.*)\s*and\s*(.*)', flags=re.IGNORECASE)
r_time_conversion = re.compile('\s*(.*)\s*(year|month|week|day|hour|minute|second|microsecond|millisecond)(?:s)?', flags=re.IGNORECASE)
r_time_conversion_to = re.compile('\s*(.*)\s*(year|month|week|day|hour|minute|second|microsecond|millisecond)(?:s)?\s*to\s*(year|month|week|day|hour|minute|second|microsecond|millisecond)(?:s)', flags=re.IGNORECASE)
r_tz_between = re.compile('(?:time difference|tz diff|time diff|time zone difference|timezone difference|timezone diff|time zone diff)?(?:s)?\s*between\s*(.*)\s*and\s*(.*)', flags=re.IGNORECASE)
r_time_plus = re.compile('(.*)\s*(?:plus|\+|after|from)\s*(.*)', flags=re.IGNORECASE)
r_time_minus = re.compile('(.*)\s*(?:minus|\-)\s*(.*)', flags=re.IGNORECASE)
r_time_before = re.compile('(.*)\s*(?:before)\s*(.*)', flags=re.IGNORECASE)
r_weekday_pre = re.compile('(?:day of week|weekday|week day|what day is|what day of week is)\s*(.*)', flags=re.IGNORECASE)
r_weekday_post = re.compile('(.*)\s*(?:day of week|weekday|week day|what day|what day of week)', flags=re.IGNORECASE)
r_workdays_since = re.compile('(?:workdays|work days)\s*since\s*(.*)', flags=re.IGNORECASE)
r_workdays_until = re.compile('(?:workdays|work days)\s*until\s*(.*)', flags=re.IGNORECASE)
r_workdays_between = re.compile('(?:workdays|work days)\s*between\s*(.*)\s*and\s*(.*)', flags=re.IGNORECASE)
r_workhours_since = re.compile('(?:workhours|work hours)\s*since\s*(.*)', flags=re.IGNORECASE)
r_workhours_until = re.compile('(?:workhours|work hours)\s*until\s*(.*)', flags=re.IGNORECASE)
r_workhours_between = re.compile('(?:workhours|work hours)\s*between\s*(.*)\s*and\s*(.*)', flags=re.IGNORECASE)
r_timezone_translation = re.compile('(.*)?\s(?:in|to)\s(.*)', flags=re.IGNORECASE)
r_timezone_translation_in_to = re.compile('(.*)(?:in)\s(.*)\s(?:to)\s(.*)', flags=re.IGNORECASE)
r_hour_minute_timezone = re.compile('((?:[0[0-9]|1[0-9]|2[0-3]):[0-5][0-9])?\s(.*)', flags=re.IGNORECASE)
r_timezone = re.compile('(.*)\s(?:timezone|timezones|tz)', flags=re.IGNORECASE)
r_calendar_year = re.compile('(?:cal year|calendar year)\s*(.*)', flags=re.IGNORECASE)
r_calendar_month = re.compile('(?:calendar|cal|month|cal month|calendar month)\s*(.*)', flags=re.IGNORECASE)
r_timezone_2 = re.compile('(?:timezone in|timezones in|tz in|timezone|timezones|tz)\s(.*)?', flags=re.IGNORECASE)
r_timezone_cities = re.compile('(?:cities with timezones|countries with timezones|cities with tz|countries with tz)\s(.*)?', flags=re.IGNORECASE)
r_weeknum = re.compile('(?:week number|week num|weeknum|week)', flags=re.IGNORECASE)


def handler_time_now_local():
    return get_local_now()


def handler_time_now_utc():
    return get_utcnow()


def dt_normalize(start_dt, end_dt) -> (datetime, datetime):
    if type(start_dt) is str:
        start_dt = dateparser_parse_dt(start_dt)
        logger.debug("Start time: {}".format(start_dt))
    if type(end_dt) is str:
        end_dt = dateparser_parse_dt(end_dt)
        logger.debug("End time: {}".format(end_dt))
    return start_dt, end_dt


def handler_tz_diff(start_tz_s: str, end_tz_s: str) -> dict:
    start_tz = resolve_timezone(start_tz_s)
    end_tz = resolve_timezone(end_tz_s)
    diff = timedelta(seconds=start_tz["tz_offset_seconds"] - end_tz["tz_offset_seconds"])
    return dict(start=start_tz,
                end=end_tz,
                diff=td_pretty(diff))


def handler_time_diff(start_dt, end_dt) -> dict:
    start_dt, end_dt = dt_normalize(start_dt, end_dt)
    diff = start_dt - end_dt
    return dict(start=dt_pretty(start_dt),
                end=dt_pretty(end_dt),
                diff=td_pretty(diff))


def handler_time_conversion(time_d, input_f, output_f=None) -> dict:
    time_d = int(time_d)
    if input_f in ['year']:
        td = timedelta(days=365*time_d)
    elif input_f in ['month']:
        td = timedelta(days=30*time_d)
    elif input_f in ['week', 'day', 'hour', 'minute', 'second', 'microsecond', 'millisecond']:
        td = timedelta(**{"{}s".format(input_f): time_d})
    else:
        td = timedelta()
    td_p = td_pretty(td)
    if not output_f:
        out = td_p['duration_human']
    else:
        out = "{} {}s".format(td_p['totals']["{}s".format(output_f)], output_f)
    return dict(out=out, timedelta=td_p)


def handler_time_since_until(start_dt_s: str) -> dict:
    return handler_time_diff(start_dt_s, get_local_now())


def handler_workdays_diff(start_dt, end_dt) -> dict:
    start_dt, end_dt = dt_normalize(start_dt, end_dt)
    diff = workday_diff(start_dt, end_dt)
    return dict(start=dt_pretty(start_dt),
                end=dt_pretty(end_dt),
                diff=td_pretty(diff))


def handler_workdays_since_until(start_dt_s: str) -> dict:
    return handler_workdays_diff(dateparser_parse_dt(start_dt_s), get_local_now())


def handler_workhours_diff(start_dt, end_dt) -> dict:
    start_dt, end_dt = dt_normalize(start_dt, end_dt)
    diff = workhours_diff(start_dt, end_dt)
    return dict(start=dt_pretty(start_dt),
                end=dt_pretty(end_dt),
                diff=td_pretty(diff))


def handler_workhours_since_until(start_dt_s: str):
    return handler_workhours_diff(dateparser_parse_dt(start_dt_s), get_local_now())


def timezone_mangle(dt_s: str, timezone_like_s: str):
    logger.debug("Timezone translation between {} | {}".format(dt_s, timezone_like_s))
    if dt_s.lower().strip() == "time":
        dt_s = "now"
    src_dt = dateparser_parse_dt(dt_s)
    logger.debug("Source time: {}".format(src_dt))
    tz = resolve_timezone(timezone_like_s, src_dt)
    logger.debug("Destination timezone: {}".format(tz))
    if not tz:
        tz, dst_dt, offset = {}, src_dt, {}
    else:
        offset = tz.get('tz_offset')
    return src_dt, offset, tz


def handler_timezone_creation(dt_s: str, timezone_like_s: str):
    src_dt, offset, tz = timezone_mangle(dt_s, timezone_like_s)
    tzinfo = tzinfo_from_offset(offset)
    dst_dt = tzinfo[0].localize(src_dt.replace(tzinfo=None))
    return src_dt, dst_dt, tz


def handler_timezone_translation(dt_s: str, timezone_like_s: str):
    src_dt, offset, tz = timezone_mangle(dt_s, timezone_like_s)
    dst_dt = dt_tz_translation(src_dt, offset)
    return src_dt, dst_dt, tz


def handler_timezone_translation_in_to(dt_s: str, timezone_like_s: str, timezone_like_d: str):
    _, src_dt, _ = handler_timezone_creation(dt_s, timezone_like_s)
    _, dst_offset, dst_tz = timezone_mangle(dt_s, timezone_like_d)
    dst_dt = dt_tz_translation(src_dt, dst_offset)
    return src_dt, dst_dt, dst_tz


def handler_generic_parser(dt_s: str) -> datetime:
    return dateparser_parse_dt(dt_s)


def handler_time_in_parser(dt_s: str) -> datetime:
    return query_to_format_result("now to {}".format(dt_s), None)


def handler_timezone(timezone_s: str):
    return resolve_timezone(timezone_s)


def get_timedelta(td: str, direction='future'):
    if direction in ['future']:
        prefix = 'in {}'
    else:
        prefix = '{} ago'
    other_dt = dateparser_parse_dt(prefix.format(td))
    now = get_local_now()
    if direction in ['future']:
        return other_dt - now
    else:
        return now - other_dt


def handler_time_plus(dt_s: str, td: str):
    dt = dateparser_parse_dt(dt_s)
    td = get_timedelta(td)
    return dt + td


def handler_time_plus_rev(td: str, dt_s: str, ):
    return handler_time_plus(dt_s, td)


def handler_time_minus(dt_s: str, td: str):
    dt = dateparser_parse_dt(dt_s)
    td = get_timedelta(td, direction='past')
    return dt - td


def handler_time_before(td: str, dt_s: str):
    dt = dateparser_parse_dt(dt_s)
    td = get_timedelta(td, direction='past')
    return dt - td


def handler_calendar(dt_s: str):
    dt = dateparser_parse_dt(dt_s)
    return {
        "month": calendar.month(int(dt.strftime("%Y")), int(dt.strftime("%m"))),
        "year": calendar.calendar(int(dt.strftime("%Y"))),
    }


QUERY_TYPE_DT_TR = "datetime_translation"
QUERY_TYPE_DT = "datetime_details"
QUERY_TYPE_TZ = "timezone"
QUERY_TYPE_TD = "timedelta"
QUERY_TYPE_CONV = "conversion"
QUERY_TYPE_CAL = "calendar"

h_default = ''
h_unix_s = 'dt->unix_s'
h_unix_ms = 'dt->unix_ms'
h_week_num = 'dt->week_number'
h_tz_offset = 'tz->tz_offset'
h_tz_iana = 'tz->found_from_offset_tzs'
h_tz_abbr = 'tz->found_from_abbr_tzs'
h_time_in = 'dt->hh:mm'
h_translation = 'dt->iso8601_full'
h_default_dt = 'dt->iso8601_full'
h_default_td = 'timedelta->diff->duration_human'
h_default_conv = 'conv->out'
h_day_of_week = 'dt->locale_day_of_week'
h_cal_year = 'cal->year'
h_cal_month = 'cal->month'

regex_handlers = [
    (r_time_conversion_to, handler_time_conversion, QUERY_TYPE_CONV, [h_default_conv]),
    (r_time_conversion, handler_time_conversion, QUERY_TYPE_CONV, [h_default_conv]),
    (r_time_in_epoch_s_now, handler_time_now_local, QUERY_TYPE_DT, [h_unix_s]),
    (r_time_in_epoch_s_now, handler_time_now_utc, QUERY_TYPE_DT, [h_unix_s]),
    (r_time_in_epoch_s2, handler_generic_parser, QUERY_TYPE_DT, [h_unix_s]),
    (r_time_in_epoch_s3, handler_generic_parser, QUERY_TYPE_DT, [h_unix_s]),
    (r_time_in_epoch_ms_now, handler_time_now_local, QUERY_TYPE_DT, [h_unix_ms]),
    (r_time_in_epoch_ms_now, handler_time_now_utc, QUERY_TYPE_DT, [h_unix_ms]),
    (r_time_in_epoch_ms2, handler_generic_parser, QUERY_TYPE_DT, [h_unix_ms]),
    (r_time_in_epoch_ms3, handler_generic_parser, QUERY_TYPE_DT, [h_unix_ms]),
    (r_timezone_translation, handler_timezone_translation, QUERY_TYPE_DT_TR, [h_translation]),
    (r_timezone_translation_in_to, handler_timezone_translation_in_to, QUERY_TYPE_DT_TR, [h_translation]),
    (r_time_since, handler_time_since_until, QUERY_TYPE_TD, [h_default_td]),
    (r_time_until, handler_time_since_until, QUERY_TYPE_TD, [h_default_td]),
    (r_time_between, handler_time_diff, QUERY_TYPE_TD, [h_default_td]),
    (r_tz_between, handler_tz_diff, QUERY_TYPE_TD, [h_default_td]),
    (r_time_plus, handler_time_plus, QUERY_TYPE_DT, [h_default_dt]),
    (r_time_plus, handler_time_plus_rev, QUERY_TYPE_DT, [h_default_dt]),
    (r_time_minus, handler_time_minus, QUERY_TYPE_DT, [h_default_dt]),
    (r_time_before, handler_time_before, QUERY_TYPE_DT, [h_default_dt]),
    (r_weekday_pre, handler_generic_parser, QUERY_TYPE_DT, [h_day_of_week]),
    (r_weekday_post, handler_generic_parser, QUERY_TYPE_DT, [h_day_of_week]),
    (r_workdays_since, handler_workdays_since_until, QUERY_TYPE_TD, [h_default_td]),
    (r_workdays_until, handler_workdays_since_until, QUERY_TYPE_TD, [h_default_td]),
    (r_workdays_between, handler_workdays_diff, QUERY_TYPE_TD, [h_default_td]),
    (r_workhours_since, handler_workhours_since_until, QUERY_TYPE_TD, [h_default_td]),
    (r_workhours_until, handler_workhours_since_until, QUERY_TYPE_TD, [h_default_td]),
    (r_workhours_between, handler_workhours_diff, QUERY_TYPE_TD, [h_default_td]),
    (r_time_in, handler_time_in_parser, QUERY_TYPE_DT, [h_time_in]),
    (r_timezone, handler_timezone, QUERY_TYPE_TZ, [h_tz_offset]),
    (r_timezone_cities, handler_timezone, QUERY_TYPE_TZ, [h_tz_iana, h_tz_abbr]),
    (r_timezone_2, handler_timezone, QUERY_TYPE_TZ, [h_tz_offset]),
    (r_hour_minute_timezone, handler_timezone_creation, QUERY_TYPE_DT_TR, [h_translation]),
    (r_calendar_year, handler_calendar, QUERY_TYPE_CAL, [h_cal_year]),
    (r_calendar_month, handler_calendar, QUERY_TYPE_CAL, [h_cal_month]),
    (r_weeknum, handler_generic_parser, QUERY_TYPE_DT, [h_week_num]),
    (r_generic, handler_generic_parser, QUERY_TYPE_DT, [h_default_dt]),
]


def try_regex(r, s):
    try:
        m = re.match(r, s)
    except:
        return None
    if m:
        groups = m.groups()
        return groups


def tokenize(s):
    solutions = []
    for r, h, t, his in regex_handlers:
        logger.debug("Trying regex: {}".format(r))
        g = try_regex(r, s)
        if g is not None:
            try:
                logger.debug("Matched regex: {}".format(r))
                logger.debug("Running handler: {} | query_type: {} | output_type: {}".format(h.__name__, t, his))
                result = h(*g)
            except Exception as e:
                logger.debug("Exception from handler: {} -> {}".format(h.__name__, e))
                continue
            if result is not None:
                logger.debug("Matched regex: {}".format(r))
                for hi in his:
                    solution = h.__name__, result, t, hi
                    if solution:
                        solutions.append(solution)
    return solutions


def pretty_print_dict(obj):
    formatted_json = json.dumps(obj, indent=2, ensure_ascii=False)
    colorful_json = highlight(formatted_json, lexers.JsonLexer(), formatters.TerminalFormatter())
    print(colorful_json)
    return colorful_json


def show_magic_results(obj, args):
    rv = []
    for solution in obj['solutions']:
        entry_proxy = Cut(solution, sep='->')
        if args.show:
            highlight_entry = args.show
        else:
            highlight_entry = solution["highlight"]
        try:
            highlight_result = entry_proxy[highlight_entry]
        except Exception as e:
            logger.debug("Exception from magic result: {} -> {}".format(highlight_entry, e))
            continue
        if args.handlers:
            # TODO: not working yet:
            # user_handlers = args.handlers.split(',')
            # for user_handler in user_handlers:
            to_print = "{} -> {}".format(solution['handler'], highlight_result)
        else:
            to_print = highlight_result
        rv.append(to_print)
        print(to_print)
        if len(rv) >= args.results:
            break
    return rv


def dt_pretty(dt):
    rv = {}
    global custom_locale
    rv["iso8601_full"] = dt.strftime(ISO_FORMAT)
    rv["iso8601_utc"] = dt_tz_translation(dt, to_tz_offset='+00:00',
                                          from_tz_offset=dt.strftime('%z')).strftime(ISO_Z_FORMAT)
    rv["iso8601_date"] = dt.strftime('%Y-%m-%d')
    rv["iso8601_time"] = dt.strftime('%H:%M:%S')
    rv["day_of_week_number"] = dt.strftime("%w")
    rv["week_number"] = dt.strftime("%W")
    rv["locale"] = custom_locale
    with setlocale(locale.LC_TIME, custom_locale.get("lc_time")):
        rv["locale_month"] = dt.strftime("%B")
        rv["locale_month_short"] = dt.strftime("%b")
        rv["locale_day_of_week_short"] = dt.strftime("%a")
        rv["locale_day_of_week"] = dt.strftime("%A")
        rv["locale_date"] = dt.strftime("%x")
        rv["locale_time"] = dt.strftime("%X")
        rv["locale_dt"] = dt.strftime("%c")
    rv["tz_offset"] = dt.strftime("%z")
    rv["tz_name"] = dt.strftime("%Z")
    rv["hh:mm"] = dt.strftime("%H:%M")
    rv["emoji_time"] = time_to_emoji(dt)
    rv["unix_s"] = get_s_since_epoch(dt)
    rv["unix_ms"] = get_ms_since_epoch(dt)
    return rv


def td_pretty(td):
    remainders = td_remainders(td)
    rv = {
        "sign": '-' if td.days > 0 else '+',
        "in_the": 'future' if td.days < 0 else 'past',
        "time_ago": time_ago(td),
        "duration_iso8601": td_iso8601(td),
        "duration_machine": td_machine(remainders),
        "duration_human": td_human(remainders),
        "duration_human_months_weeks": td_human_months_weeks(remainders),
        "totals": td_totals(td),
        "remainders": remainders,
    }
    return rv


def resolve_query_type(query):
    solutions = tokenize(query)
    if not solutions:
        dt = get_local_now()
        return [["now", dt, QUERY_TYPE_DT, h_default_dt]]
    return solutions


def resolve_query(query, allowed_queries=None):
    rv = {
        "query": query,
        "solutions": [],
    }
    solutions = resolve_query_type(query)
    if not allowed_queries:
        allowed_queries = [QUERY_TYPE_DT, QUERY_TYPE_DT_TR, QUERY_TYPE_TD, QUERY_TYPE_TZ, QUERY_TYPE_CAL, QUERY_TYPE_CONV]
    for sol_id, solution in enumerate(solutions):
        element = {}
        handler, results, query_type, hi = solution
        element["handler"] = handler
        element["query_type"] = query_type
        element["highlight"] = hi
        element["solution_id"] = sol_id
        try:
            if query_type not in allowed_queries:
                continue
            if query_type == QUERY_TYPE_DT:
                element["dt"] = dt_pretty(results)
            elif query_type == QUERY_TYPE_DT_TR:
                element["src_dt"] = dt_pretty(results[0])
                element["dt"] = dt_pretty(results[1])
                element["tz"] = results[2]
            elif query_type == QUERY_TYPE_TZ:
                element["tz"] = results
            elif query_type == QUERY_TYPE_TD:
                element["timedelta"] = results
            elif query_type == QUERY_TYPE_CAL:
                element["cal"] = results
            elif query_type == QUERY_TYPE_CONV:
                element["conv"] = results
            rv["solutions"].append(element)
        except Exception as e:
            continue
    return rv


def test():
    test_strings = [
        None,
        "",
        "s",
        "    ",
        "Time  since 2019-05-12",
        "Since yesterday",
        "time between yesterday and tomorrow",
        "time until 25 december",
        "time sinc",
        "now in milliseconds",
        "seconds since epoch",
        "1992-01-27 to epoch",
        "milliseconds since 1992-01-27",
        "now in sofia",
        "now in PST",
        "2 hours ago to Sydney",
        "now in +03:00",
        "now in dublin",
        "workdays since 2/07/2020 12:00",
        "workhours since 2/07/2020 12:00",
    ]
    for s in test_strings:
        print("{} -> {}".format(s, resolve_query(s)))


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('query', nargs='*', default="", help="freeform")
    parser.add_argument('--locale', dest='locale')
    parser.add_argument('--tz', dest='tz', default='local')
    parser.add_argument('--handlers', dest='handlers', action='store_true')
    parser.add_argument('--results', dest='results', type=int, default=10)
    parser.add_argument('--show', dest='show')
    parser.add_argument('--full', dest='full', action='store_true')
    parser.add_argument('--debug', dest='debug', action='store_true')
    args = parser.parse_args()
    return args


def setup_logging_level(debug=False):
    log_level = logging.DEBUG if debug else logging.ERROR
    logger.setLevel(log_level)
    logger.debug("Debugging enabled")


def main(args):
    global custom_locale
    global custom_tz
    custom_locale = resolve_locale(args.locale)
    custom_tz = resolve_timezone(args.tz)
    if custom_locale:
        logger.debug("Locale understood as: {}".format(custom_locale))
    query = ' '.join(args.query)
    result = resolve_query(query)
    if args.full:
        pretty_print_dict(result)
    return show_magic_results(result, args)

